Imports Microsoft.VisualBasic
Imports System
Namespace RestrictedBoltzmann
    Friend Class RestrictedBoltzmannProgram
        Shared Sub Main(ByVal args() As String)
            Console.WriteLine(vbLf & "Begin Restricted Boltzmann Machine demo " & vbLf)

            Console.WriteLine("Films: Alien, Inception, Spy, EuroTrip,")
            Console.WriteLine(" Gladiator, Spartacus " & vbLf)

            ' type A person = likes only Science Fiction films
            ' type B person = likes only Comedy
            ' type C person = likes only History

            Dim trainData(11)() As Integer
            trainData(0) = New Integer() {1, 1, 0, 0, 0, 0} ' A
            trainData(1) = New Integer() {0, 0, 1, 1, 0, 0} ' B
            trainData(2) = New Integer() {0, 0, 0, 0, 1, 1} ' C

            trainData(3) = New Integer() {1, 1, 0, 0, 0, 1} ' noisy A
            trainData(4) = New Integer() {0, 0, 1, 1, 0, 0} ' B
            trainData(5) = New Integer() {0, 0, 0, 0, 1, 1} ' C

            trainData(6) = New Integer() {1, 0, 0, 0, 0, 0} ' weak A
            trainData(7) = New Integer() {0, 0, 1, 0, 0, 0} ' weak B
            trainData(8) = New Integer() {0, 0, 0, 0, 1, 0} ' weak C

            trainData(9) = New Integer() {1, 1, 0, 1, 0, 0} ' noisy A
            trainData(10) = New Integer() {1, 0, 1, 1, 0, 0} ' noisy B
            trainData(11) = New Integer() {0, 0, 1, 0, 1, 1} ' noisy C

            Console.WriteLine("Film like-dislike data is: " & vbLf)
            ShowData(trainData, 5, True)

            Dim numVisible As Integer = 6
            Dim numHidden As Integer = 3
            Console.WriteLine("Creating a RBM ")
            Console.WriteLine("Setting numVisible = " & numVisible)
            Console.WriteLine("Setting numHidden  = " & numHidden & vbLf)

            Dim rbm As New Machine(numVisible, numHidden)

            Dim learnRate As Double = 0.01
            Dim maxEpochs As Integer = 1000
            Console.WriteLine("Training RBM using CD1 algorithm ")
            Console.WriteLine("Setting learnRate = " & learnRate.ToString("F3"))
            Console.WriteLine("Setting maxEpochs = " & maxEpochs)
            rbm.Train(trainData, learnRate, maxEpochs)
            Console.WriteLine("Training complete " & vbLf)

            Console.WriteLine("Trained machine's weights and biases are: " & vbLf)
            rbm.Dump(False, True, True)

            Console.WriteLine("Using trained machine . . . " & vbLf)

            Dim visibles() As Integer = {1, 1, 0, 0, 0, 0}
            Dim computedHidden() As Integer = rbm.HiddenFromVis(visibles)
            Console.Write("visible = ")
            ShowVector(visibles, False)
            Console.Write("  ->  ")
            ShowVector(computedHidden, True)

            Dim hiddens() As Integer = {1, 1, 0}
            Dim computedVisibles() As Integer = rbm.VisibleFromHid(hiddens)
            Console.Write("hidden = ")
            ShowVector(hiddens, False)
            Console.Write("  ->  ")
            ShowVector(computedVisibles, True)

            Console.WriteLine(vbLf & "End RBM demo " & vbLf)
            Console.ReadLine()
        End Sub ' Main

        Private Shared Sub ShowData(ByVal data()() As Integer, ByVal numRows As Integer, ByVal showLast As Boolean)
            For i = 0 To numRows - 1
                Console.Write("[" & i.ToString().PadLeft(2) & "]   ")
                For j = 0 To data(i).Length - 1
                    Console.Write(data(i)(j) & "  ")
                Next j
                Console.WriteLine("")
            Next i
            If showLast = True Then
                Console.WriteLine(" . . . ")
                Console.Write("[" & (data.Length - 1).ToString().PadLeft(2) & "]   ")
                For j = 0 To data(data.Length - 1).Length - 1
                    Console.Write(data(data.Length - 1)(j) & "  ")
                Next j
                Console.WriteLine("")
            End If
            Console.WriteLine("")
        End Sub

        Private Shared Sub ShowVector(ByVal vector() As Integer, ByVal newline As Boolean)
            For i = 0 To vector.Length - 1
                Console.Write(vector(i) & " ")
            Next i
            If newline = True Then
                Console.WriteLine("")
            End If
        End Sub

    End Class ' Program

    Public Class Machine
        Public rnd As Random

        Public numVisible As Integer
        Public numHidden As Integer

        Public visValues() As Integer ' visible node values (0, 1)
        Public visProbs() As Double
        Public visBiases() As Double

        Public hidValues() As Integer
        Public hidProbs() As Double
        Public hidBiases() As Double

        Public vhWeights()() As Double

        Public Sub New(ByVal numVisible As Integer, ByVal numHidden As Integer)
            Me.rnd = New Random(0)

            Me.numVisible = numVisible
            Me.numHidden = numHidden

            ' allocate arrays & the weights matrix
            visValues = New Integer(numVisible - 1) {}
            visProbs = New Double(numVisible - 1) {}
            visBiases = New Double(numVisible - 1) {}

            hidValues = New Integer(numHidden - 1) {}
            hidProbs = New Double(numHidden - 1) {}
            hidBiases = New Double(numHidden - 1) {}

            vhWeights = New Double(numVisible - 1)() {} ' visible-to-hidden
            For i = 0 To numVisible - 1
                vhWeights(i) = New Double(numHidden - 1) {}
            Next i

            ' small random values for initial weights & biases
            Dim lo As Double = -0.4
            Dim hi As Double = +0.4
            For i = 0 To numVisible - 1
                For j = 0 To numHidden - 1
                    vhWeights(i)(j) = (hi - lo) * rnd.NextDouble + lo
                Next j
            Next i

            For i = 0 To numVisible - 1
                visBiases(i) = (hi - lo) * rnd.NextDouble + lo
            Next i

            For i = 0 To numHidden - 1
                hidBiases(i) = (hi - lo) * rnd.NextDouble + lo
            Next i
        End Sub ' ctor

        Public Sub Train(ByVal trainData()() As Integer, ByVal learnRate As Double, ByVal maxEpochs As Integer)
            Dim indices(trainData.Length - 1) As Integer
            For i = 0 To indices.Length - 1
                indices(i) = i
            Next i

            Dim epoch As Integer = 0
            Do While epoch < maxEpochs
                Shuffle(indices)

                For idx = 0 To indices.Length - 1 ' each data item
                    Dim i As Integer = indices(idx) ' i points to curr train data

                    ' copy visible values from train data into Machine
                    For j = 0 To numVisible - 1
                        visValues(j) = trainData(i)(j)
                    Next j

                    ' compute hidden node values ('h' in Wikipedia)
                    For h = 0 To numHidden - 1
                        Dim sum As Double = 0.0
                        For v = 0 To numVisible - 1
                            sum += visValues(v) * vhWeights(v)(h)
                        Next v

                        sum += hidBiases(h) ' add the hidden bias
                        hidProbs(h) = LogSig(sum) ' compute prob of h activation
                        Dim pr As Double = rnd.NextDouble ' determine 0/1 h node value
                        If hidProbs(h) > pr Then
                            hidValues(h) = 1
                        Else
                            hidValues(h) = 0
                        End If
                    Next h

                    ' compute positive gradient =  outer product of v & h
                    Dim posGrad()() As Integer = OuterProd(visValues, hidValues)

                    ' reconstruct visual Nodes as v'
                    Dim vPrime(numVisible - 1) As Integer ' v' in Wikipedia
                    For v = 0 To numVisible - 1
                        Dim sum As Double = 0.0
                        For h = 0 To numHidden - 1
                            sum += hidValues(h) * vhWeights(v)(h)
                        Next h
                        sum += visBiases(v) ' add visible bias
                        Dim probActiv As Double = LogSig(sum)
                        Dim pr As Double = rnd.NextDouble
                        If probActiv > pr Then
                            vPrime(v) = 1
                        Else
                            vPrime(v) = 0
                        End If
                    Next v

                    ' compute new hidden Nodes as h', using v'
                    Dim hPrime(numHidden - 1) As Integer
                    For h = 0 To numHidden - 1
                        Dim sum As Double = 0.0
                        For v = 0 To numVisible - 1
                            sum += vPrime(v) * vhWeights(v)(h)
                        Next v
                        sum += hidBiases(h) ' add the hidden bias
                        Dim probActiv As Double = LogSig(sum) ' apply activation
                        Dim pr As Double = rnd.NextDouble ' determine 0/1 node value
                        If probActiv > pr Then
                            hPrime(h) = 1
                        Else
                            hPrime(h) = 0
                        End If
                    Next h

                    ' compute negative grad using v' and h'
                    Dim negGrad()() As Integer = OuterProd(vPrime, hPrime)

                    ' update weights
                    For row = 0 To numVisible - 1
                        For col = 0 To numHidden - 1
                            vhWeights(row)(col) += learnRate * (posGrad(row)(col) - negGrad(row)(col))
                        Next col
                    Next row

                    ' update visBiases
                    For v = 0 To numVisible - 1
                        visBiases(v) += learnRate * (visValues(v) - vPrime(v))
                    Next v
                    ' update hidBiases
                    For h = 0 To numHidden - 1
                        hidBiases(h) += learnRate * (hidValues(h) - hPrime(h))
                    Next h

                Next idx ' for-each train data

                epoch += 1
            Loop ' while
        End Sub ' Train

        Public Shared Function OuterProd(ByVal visValues() As Integer, ByVal hidValues() As Integer) As Integer()()
            Dim rows As Integer = visValues.Length
            Dim cols As Integer = hidValues.Length
            Dim result(rows - 1)() As Integer
            For i = 0 To rows - 1
                result(i) = New Integer(cols - 1) {}
            Next i

            For i = 0 To rows - 1
                For j = 0 To cols - 1
                    result(i)(j) = visValues(i) * hidValues(j)
                Next j
            Next i

            Return result
        End Function

        Public Function LogSig(ByVal x As Double) As Double
            If x < -20.0 Then
                Return 0.0000000001
            ElseIf x > 20.0 Then
                Return 0.9999999999
            Else
                Return 1.0 / (1.0 + Math.Exp(-x))
            End If
        End Function

        Public Sub Shuffle(ByVal indices() As Integer)
            For i = 0 To indices.Length - 1
                Dim ri As Integer = rnd.Next(i, indices.Length)
                Dim tmp As Integer = indices(i)
                indices(i) = indices(ri)
                indices(ri) = tmp
            Next i
        End Sub

        Public Function HiddenFromVis(ByVal visibles() As Integer) As Integer()
            Dim result(numHidden - 1) As Integer

            For h = 0 To numHidden - 1
                Dim sum As Double = 0.0
                For v = 0 To numVisible - 1
                    sum += visibles(v) * vhWeights(v)(h)
                Next v

                sum += hidBiases(h) ' add the hidden bias
                Dim probActiv As Double = LogSig(sum) ' compute prob of h activation
                ' Console.WriteLine("Hidden [" + h + "] activation probability = " + probActiv.ToString("F4"));
                Dim pr As Double = rnd.NextDouble ' determine 0/1 h node value
                If probActiv > pr Then
                    result(h) = 1
                Else
                    result(h) = 0
                End If
            Next h
            Return result
        End Function

        Public Function VisibleFromHid(ByVal hiddens() As Integer) As Integer()
            Dim result(numVisible - 1) As Integer

            For v = 0 To numVisible - 1
                Dim sum As Double = 0.0
                For h = 0 To numHidden - 1
                    sum += hiddens(h) * vhWeights(v)(h)
                Next h
                sum += visBiases(v) ' add visible bias
                Dim probActiv As Double = LogSig(sum)
                ' Console.WriteLine("Visible [" + v + "] activation probability = " + probActiv.ToString("F4"));
                Dim pr As Double = rnd.NextDouble
                If probActiv > pr Then
                    result(v) = 1
                Else
                    result(v) = 0
                End If
            Next v
            Return result
        End Function

        Public Sub Dump(ByVal showValues As Boolean, ByVal showWeights As Boolean, ByVal showBiases As Boolean)
            If showValues = True Then
                For i = 0 To numVisible - 1
                    Console.Write("visible node [" & i & "] value = " & visValues(i))
                    Console.WriteLine("  prob = " & visProbs(i).ToString("F4"))
                Next i
                Console.WriteLine("")

                For j = 0 To numHidden - 1
                    Console.Write("hidden node [" & j & "] value = " & hidValues(j))
                    Console.WriteLine("  prob = " & hidProbs(j).ToString("F4"))
                Next j
                Console.WriteLine("")
            End If

            If showWeights = True Then
                For i = 0 To numVisible - 1
                    For j = 0 To numHidden - 1
                        Dim x As Double = vhWeights(i)(j)
                        If x >= 0.0 Then
                            Console.Write(" ")
                        End If
                        Console.Write(vhWeights(i)(j).ToString("F4") & "  ")
                    Next j
                    Console.WriteLine("")
                Next i
                Console.WriteLine("")
            End If

            If showBiases = True Then
                For i = 0 To numVisible - 1
                    Console.WriteLine("visible bias [" & i & "] value = " & visBiases(i).ToString("F4"))
                Next i
                Console.WriteLine("")

                For j = 0 To numHidden - 1
                    Console.WriteLine("hidden bias [" & j & "] value = " & hidBiases(j).ToString("F4"))
                Next j
                Console.WriteLine("")
            End If
        End Sub ' Dump

    End Class ' Boltzmann Machine

End Namespace ' ns